@@ -73,7 +73,7 @@ use Data::Dumper;
     );
 warn Dumper $results;
 
-    my $results = $e->cluster->health(
+    $results = $e->cluster->health(
         index => $index_name,
     );
 warn Dumper $results;
\ No newline at end of file
