@@ -1,7 +1,8 @@
 #!perl -w
 use strict;
-use AnyEvent;
+#use AnyEvent;
 use Search::Elasticsearch::Async;
+use Search::Elasticsearch::Async::Bulk;
 use Promises qw[collect deferred];
 
 use Getopt::Long;
@@ -20,8 +21,9 @@ use Dancer::SearchApp::IndexSchema qw(create_mapping find_or_create_index %indic
 use Dancer::SearchApp::Utils qw(await);
 use Dancer::SearchApp::Extractor;
 
-#use lib 'C:/Users/Corion/Projekte/Apache-Tika/lib';
-use CORION::Apache::Tika::Server;
+use lib 'C:/Users/Corion/Projekte/Apache-Tika-Async/lib';
+use Apache::Tika::Server;
+use Dancer::SearchApp::HTMLSnippet;
 
 use JSON::MaybeXS;
 my $true = JSON->true;
@@ -32,17 +34,17 @@ my $false = JSON->false;
   # index a directory and its subdirectories
   index-filesytem.pl $HOME
   
-  # Use defaults from ./fs-import.yml
+  # Use settings from ~/myconfig.yml
   index-filesystem.pl -c ~/myconfig.yml
 
   # Drop and recreate index:
-  index-filesystem.pl -f ./fs-import.yml
+  index-filesystem.pl -f -c ./fs-import.yml
 
 =cut
 
 GetOptions(
     'force|f' => \my $force_rebuild,
-    'config|c' => \my $config_file,
+    'config|c:s' => \my $config_file,
     # How can we easily pass the options for below as command line parameters?!
 );
 $config_file ||= 'fs-import.yml';
@@ -52,10 +54,11 @@ my $file_config = LoadFile($config_file);
 my $config = get_defaults(
     env      => \%ENV,
     config   => $file_config,
-    #defaults => \%
+    #defaults => \%defaults,
     names => [
         ['elastic_search/index' => 'elastic_search/index' => 'SEARCHAPP_ES_INDEX', 'dancer-searchapp'],
         ['elastic_search/nodes' => 'elastic_search/nodes' => 'SEARCHAPP_ES_NODES', 'localhost:9200'],
+        ['fs' => 'fs' => undef, []],
     ],
 );
 
@@ -65,31 +68,34 @@ my $e = Search::Elasticsearch::Async->new(
     nodes => [
         $config->{elastic_search}->{nodes},
     ],
-    plugins => ['Langdetect'],
+    #plugins => ['Langdetect'],
 );
 
 my $extractor = 'Dancer::SearchApp::Extractor';
 
-my $tika_glob = 'C:/Users/Corion/Projekte/Apache-Tika/jar/tika-server-*.jar';
+my $tika_glob = 'jar/tika-server-*.jar';
 my $tika_path = (sort { my $ad; $a =~ /server-1.(\d+)/ and $ad=$1;
                 my $bd; $b =~ /server-1.(\d+)/ and $bd=$1;
                 $bd <=> $ad
               } glob $tika_glob)[0];
-die "Tika not found in '$tika_glob'" unless -f $tika_path; 
-#warn "Using '$tika_path'";
-my $tika= CORION::Apache::Tika::Server->new(
+die "Tika not found in '$tika_glob'. Did you download it from https://tika.apache.org/download.html?"
+    unless -f $tika_path; 
+my $tika= Apache::Tika::Server->new(
     jarfile => $tika_path,
 );
 $tika->launch;
+warn "Launched tika";
 
-my $ok = AnyEvent->condvar;
-my $info = await $e->cat->plugins;
+#my $ok = AnyEvent->condvar;
+#warn "Requesting ES plugins";
+#my $info = await $e->cat->plugins;
 
 # Koennen wir ElasticSearch langdetect als Fallback nehmen?
-my $have_langdetect = $info =~ /langdetect/i;
-if( ! $have_langdetect ) {
-    warn "Language detection disabled";
-};
+#my $have_langdetect = $info =~ /langdetect/i;
+my $have_langdetect = 0;
+#if( ! $have_langdetect ) {
+#    warn "Language detection disabled";
+#};
 
 # https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-lang-analyzer.html
 
@@ -138,13 +144,15 @@ sub fs_recurse {
     my @folders;
 
     for my $folderspec (@{$config->{directories}}) {
-        if( ! exists $folderspec->{exclude} ) {
-            # By default, exclude hidden files
-            $folderspec->{exclude} = [qr/^\./];
-        };
+        
+        #if( ! exists $folderspec->{exclude} ) {
+        #    # By default, exclude hidden files
+        #    $folderspec->{exclude} = [qr/^\./];
+        #};
         if( ! ref $folderspec ) {
             # plain name, use this folder
             push @folders, dir($folderspec)
+            
         } else {
             my $dir = dir($folderspec->{folder});
             push @folders, $dir;
@@ -176,7 +184,10 @@ sub get_entries_from_folder {
         warn "Skipped $folder, no permissions\n";
     };
     
-    return grep { !$_->is_dir and ! /^\./ } @directories;
+    return
+        grep {! in_exclude_list($_, $config->{fs}->{exclude_files} ) }
+        grep { !$_->is_dir and ! /^\./ }
+        @directories;
 };
 
 
@@ -187,6 +198,8 @@ sub get_file_info {
     $res{ folder } = "" . $file->dir;
     $res{ folder } =~ s![\\/ ]! !g;
     
+    print "$file\n";
+    my $info;
     eval {
         $info = $tika->get_all( $file );
     };
@@ -204,6 +217,7 @@ sub get_file_info {
         my @info = await $extractor->examine(
               url => $url,
               info => $info,
+              meta => $meta,
               #content => \$content, # if we have it
               filename => $file, # if we have it
               folder => $res{ folder }, # if we have it
@@ -219,24 +233,25 @@ sub get_file_info {
         } else {
             
             # Just use what Tika found
+            
+            my $c = $info->content;
+            my $r = Dancer::SearchApp::HTMLSnippet->cleanup_tika( $c );
 
-            use HTML::Restricted;            
-            my $p = HTML::Restricted->new();
-            my $r = $p->filter( $info->content );
-
-            $res{ title } = $meta->{"meta:title"} || $file->basename;
+            $res{ title } = $meta->{"dc:title"} || $meta->{"title"} || $file->basename;
             $res{ author } = $meta->{"meta:author"}; # as HTML
             $res{ language } = $meta->{"meta:language"};
-            $res{ content } = $r->as_HTML; # as HTML
+            $res{ content } = $r; # as HTML
+            $res{ mime_type } = $meta->{"Content-Type"};
         }
     }
-    
+
     my $ctime = (stat $file)[10];
     $res{ creation_date } = strftime('%Y-%m-%d %H:%M:%S', localtime($ctime));
+    $res{ url } ||= "$file";
     \%res
 }
 
-my $ld = $e->langdetect;
+my $ld;# = $e->langdetect;
 sub detect_language {
     my( $content, $meta ) = @_;
     my $res;
@@ -267,7 +282,7 @@ if( @ARGV) {
     $config->{fs}->{directories} = [@ARGV];
 };
 
-if( ! @ARGV ) {
+if( ! @ARGV and ! @{ $config->{fs}->{directories} }) {
     # If we don't know better, scan the (complete) profile
     my $userhome = $ENV{USERPROFILE} || $ENV{HOME};
     $config->{fs}->{directories} = [{ folder => $userhome, recurse => 1 }];
@@ -275,61 +290,90 @@ if( ! @ARGV ) {
 
 my @folders = fs_recurse(undef, $config->{fs});
 for my $folder (@folders) {
-    my @entries;
-    print "Reading $folder\n";
-    push @entries, map {
-        # analyze file
-        # recurse into file parts for (ZIP) archives?!
-        # SHA1
-        get_file_info($_)
-    } get_entries_from_folder( $folder );
 
-    my $done = AnyEvent->condvar;
+    print "Reading $folder\n";
+    # We need to make this promises-based/asynchronous too so
+    # that we don't accumulate a lot of data client-side
+    my @entries = get_entries_from_folder( $folder );
+
+    my $bulk = $e->bulk_helper(
+        max_count => 10,
+        on_error => sub {
+            my($name,$data,$code) = @_;
+            warn "ES Error: $name ($code): " . Dumper $data;
+        }
+    );
 
     # Importieren
     print sprintf "Importing %d files\n", 0+@entries;
-    collect(
-        map {
-            my $msg = $_;
-            my $body = $msg->{content};
-            
-            # Stringify some fields that are prone to be objects:
-            for(qw(file url)) {
-                if( $msg->{$_} ) {
-                    $msg->{ $_} = "$msg->{$_}";
-                };
-            };
-            
-            my $lang = detect_language($body, $msg);
-            
-            $lang->then(sub{
-                my $found_lang = $_[0]; #'en';
-                #warn "Have language '$found_lang'";
-                return find_or_create_index($e, $index_name,$found_lang, 'file')
-            })
-            ->then( sub {
-                my( $full_name ) = @_;
-                #warn $msg->{mime_type};
-                # https://www.elastic.co/guide/en/elasticsearch/guide/current/one-lang-docs.html
-                #warn "Storing document into $full_name";
-                $e->index({
-                        index   => $full_name,
-                        type    => 'file', # or 'attachment' ?!
-                        id      => $msg->{url}, # we want to overwrite
-                        # index bcc, cc, to, from
-                        body    => $msg # "body" for non-bulk, "source" for bulk ...
-                        #source    => $msg
-                 });
-               })->then(sub{
-                   #warn "Done."
-               }, sub {warn $_ for @_ });
-       } @entries
-    )->then(sub {
-        print "$folder done\n";
-        $done->send;
-    });
     
-    $done->recv;
-    #$importer->flush;
+    # Process in a batch size of 10, to debug memory consumption
+    while( my @batch = splice @entries, 0, 100 ) {
+
+        await collect(
+            map {
+                # One day, this will be a Promise too
+                my $msg = get_file_info($_);
+                
+                my $body = $msg->{content};
+                
+                # Stringify some fields that are prone to be objects:
+                for(qw(file url)) {
+                    if( $msg->{$_} ) {
+                        $msg->{ $_} = "$msg->{$_}";
+                    };
+                };
+                
+                my $lang = detect_language($body, $msg);
+                
+                $lang->then(sub{
+                    my $found_lang = $_[0]; #'en';
+                    #warn "Have language '$found_lang'";
+                    return find_or_create_index($e, $index_name,$found_lang, 'file')
+                })->then( sub {
+                    my( $full_name ) = @_;
+                    #warn $msg->{mime_type};
+                    
+                    # munge the title so we get magic completion for document titles:
+                    # This should be mostly done in an Elasticsearch filter+analyzer combo
+                    # Except for bands/song titles, which we want to manually munge
+                    my @parts = map {lc $_}
+                                ((split /\s+/, $msg->{title}),
+                                (split m![\\/]!, $msg->{url}));
+                    $msg->{title_suggest} = {
+                        input => \@parts,
+                        #output => $msg->{title},
+                        
+                        # Maybe some payload to directly link to the document. Later
+                        #payload => {
+                        #        url => $msg->{url}
+                        #        # , $msg->{mime_type}
+                        #    },
+                    };
+                    
+                    # https://www.elastic.co/guide/en/elasticsearch/guide/current/one-lang-docs.html
+                    #warn "Storing document into $full_name";
+                    
+                    # Switch this to a bulk converter
+                    #$e->index({
+                    $bulk->index({
+                            index   => $full_name,
+                            type    => 'file', # or 'attachment' ?!
+                            id      => $msg->{url}, # we want to overwrite
+                            # index bcc, cc, to, from
+                            #body    => $msg # "body" for non-bulk, "source" for bulk ...
+                            source  => $msg # "body" for non-bulk, "source" for bulk ...
+                    });
+                })->then(sub{
+                       # Also add the document to the potential keywords for suggestion
+                       #warn "Done."
+                       return ()
+                })->catch(sub {undef $msg; warn $_ for @_ });
+           } @batch
+        );
+    };
+    await $bulk->flush;
+    sleep 1;
+    
+    print "$folder done\n";
 };
-#$importer->flush;
